#ifndef lint
static const char rcsid[] = 
	"@(#)$Id: rk.c,v 1.8 1994/11/02 22:37:48 schwartz Exp $";
#endif

/*
 *  RK Predictive Magic. 
 *  based on reactive.c, with substantial modifications by Scott Schwartz
 * 

 .SH "COPYRIGHT AND DISTRIBUTION"
 Copyright 1990, John J. Darragh, Ian H. Witten and Mark L. James.  Permission
 is granted to make and distribute copies of this software, provided this
 disclaimer and copyright notice are preserved on all copies.  The software may
 not, however, be sold or distributed for profit, or included with other
 software which is sold or distributed for profit, without the permission of the
 authors.
 .PP
 The Reactive Keyboard is free.  There is no shareware fee. We encourage you to
 make and distribute as many copies as you wish, for whomever you wish, as long
 as it is not for profit.  The current versions are available electronically via
 anonymous ftp from cs-sun-fsa.UCalgary.CA (internet 136.159.2.1).  Versions are
 available for Unix, IBM PC, and Macintosh.
 .PP
 It is provided in source form so that it can be modified to
 suit different people's requirements.  If you do modify it, we'd like to hear
 about it so that others will be able to benefit from your efforts.
 .PP
 \fIRk\fP is a demonstration program, not a commercial product.  We
 cannot offer services such as telephone support, mailing lists, or upgrade
 services.

 */
/*
    See also:
    Darragh and Witten, The Reactive Keyboard, Cambridge University Press,
    1992, ISBN 0-521-40375-8, QA76.9.H45D37
 */

#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <errno.h>

#ifdef NEED_STRERROR
const char *strerror (int);
#endif

#include "rk.h"

static char first_pred(struct rk *rk, context_vec_t buf);
static char* last_n_of_2 (char* out, int n, const char* a, const char* b);
static int build_menu (struct rk *rk);
static int move_up (struct rk *rk, int c);
static tree_node_t * create_node (struct rk *rk);
static void scan_up (struct rk *rk, context_vec_t theList, int c);

static const char zero_freq_template[MAX_SET] = {
  '\n',   ' ',   'e',   's',   't',   'r',   'a',   'l',
   'n',   '.',   'c',   'i',   'm',   'o',   'd',   'h',
   'p',   'u',   'f',   'b',   'w',   'g',   '-',   'y',
   '/',   'v',   'k',   '*',   'x',   '5',   '1',   '2',
   '4',   '>',   'q', '\07',   '3',   'z',   '0',   'j',
   'M',   'I',   '6',  '\'',   'E',   '~',   'A',   ',',
   'S',   'D',   'R',   '?',   'C',   '|',   'B',   'T',
   'P',   'U',   '!',   '_',   '<',   'N',   '8',   '@',
   '&',   '7',  '\\',   '"',   'J',   ')',   '(',   '[',
   ']',   'F',   'L',   ':',   'O',   'K',   '9',   'H',
   '+',   '$',   'W',   'Y',   '=',   'G',   ';',   '^',
   '{',   'X',   'Q',   '#',   '}',   'V',   '%',   'Z',
   '`',  '\b',  '\t', '\26', '\00', '\01', '\02', '\03',
 '\04', '\05', '\06', '\13',  '\f',  '\r', '\16', '\17',
 '\20', '\21', '\22', '\23', '\24', '\25', '\27', '\30',
 '\31', '\32', '\33', '\34', '\35', '\36', '\37','\177',
};


struct rk *rk_alloc (void * (*alloc)())
{
    struct rk *rk = alloc (sizeof (struct rk));
    if (!rk)
	return 0;

    memset ((char*)rk->Buf, 0, sizeof(rk->Buf));
    memset ((char*)rk->CBuf, 0, sizeof(rk->CBuf));

    rk->context[0] = rk->old_context[0] = '\n';
    rk->context[1] = rk->old_context[1] = '\0';

    rk->free_nodes = rk->root = 0;
    rk->psize = rk->next_free = 0;

    rk->AllocationCount = rk->CharsInModel = rk->FrequencySplits = 0;

    rk->memoryIsFull = 0;
    rk->pred_number = 0;
    rk->node_limit = 0;

    rk->alloc = alloc;
    memset (rk->pred_buff, 0, sizeof(rk->pred_buff));

    rk->max_freq = USHRT_MAX - 1,
    rk->max_len = 80;		/* assert < MAX_PREDICTION_LENGTH */
    rk->maxk = 12;		/* assert < TOP_K */
    rk->mem_limit = 4096;	/* Kbytes */

    return rk;
}

/* =======================================================================
 * Name - build_menu
 *
 * Purpose - Fill out prediction in pred_buff
 *
 * Arguments:
 *
 * Returns     function return --
 *
 *========================================================================
 */
static int build_menu (struct rk *rk)
{				/* fill out prediction in s */
    int i, length;
    context_vec_t tbuf;
    char *bptr = &rk->pred_buff[0];
    char c;

    length = rk->maxk;
    for (i = 0; i <= length; i++)
	tbuf[i] = rk->CBuf[i];

    length = rk->max_len;
    c = rk->first[rk->pred_number];
    for (i = 1; i <= length; i++) {
	*bptr++ = c;
	scan_up(rk, tbuf, c);
	c = first_pred(rk, tbuf);
    }
    *bptr = '\0';
    return 0;
}

/* =======================================================================
 * Name - create_node
 *
 * Purpose - Allocate new node, using our shortcut heap if possible
 *
 * Arguments:
 *
 * Returns     function return -- new node
 *
 *========================================================================
 */
static tree_node_t * create_node (struct rk *rk)
{				/* return ptr to a new node or abort */
    rk->AllocationCount ++;
    if (rk->AllocationCount > rk->node_limit) {
	rk->AllocationCount--;	/* don't lie to show_used_nodes */
	rk->memoryIsFull = 1;
	return 0;
    }

#ifndef POOL_ALLOC
    /* do this instead of pool allocation so we can use gc_malloc */
    return (rk->alloc)(sizeof(tree_node_t));
#else
    if (rk->next_free >= NODE_ALLOCATION) {
	rk->free_nodes = 
	    (tree_node_t *)(rk->alloc)(NODE_ALLOCATION * sizeof(tree_node_t));
	if (rk->free_nodes == 0) {
	    /* This is really too late, as it will screw up file_completion,
	       command_completion, and who knows what else.  But they can
	       always use -n to keep from getting here... */
	    rk->memoryIsFull = 1;
	    return 0;
	}
	rk->next_free = 0;
    }
    return &rk->free_nodes[rk->next_free++];
#endif
}

/* =======================================================================
 * Name - find_first
 *
 * Purpose - find 1st char of all pred in context
 *           [Calculate probabilities for each possible character]
 *
 * Arguments:
 *
 * Returns     function return --
 *
 *========================================================================
 */
int find_first (struct rk *rk)
{
    int i, order = 0; 
    char *p;
    tree_node_t *xptr; 
    tree_node_t **buf;

    if (!rk)
	return 0;

    buf = rk->CBuf;

    for (p = &rk->pred_set[0]; p < &rk->pred_set[MAX_SET]; p++)
	*p = 0;
    for (i = rk->maxk - 1; i >= 0; i--) {
	if (buf[i] != 0 && buf[i]->up != 0) {
	    xptr = buf[i]->up;
	    while (xptr != 0) {
		if (rk->pred_set[(int)xptr->value] == 0) {
		    rk->pred_set[(int)xptr->value]++;
		    rk->first[order++] = xptr->value;
		}
		xptr = xptr->next;
	    }
	}
    }
    for (p = &rk->zero_freq[0]; p < &rk->zero_freq[MAX_SET]; p++)
	if (rk->pred_set[(int)*p] == 0)
	    rk->first[order++] = *p;
    
    return 1;
}

/* =======================================================================
 * Name - first_pred
 *
 * Purpose -
 *
 * Arguments:
 *
 * Returns     function return --
 *
 *========================================================================
 */
static char first_pred(struct rk *rk, context_vec_t buf)
{
    int i = rk->maxk - 1;

    for (;;) {
	if (buf[i] != 0 && buf[i]->up != 0)
	    return buf[i]->up->value;
	if (i == 0)
	    return '\0';
	i--;
    }
}

/* =======================================================================
 * Name - init_reactive
 *
 * Purpose - initialize the predictor, esp. from a file
 *
 * Arguments:  u -- load factor to prejudice amount to read from file
 *
 * Returns     function return --
 *
 *========================================================================
 */
int init_reactive (struct rk *rk, const char* zero_vec, const char* prime_vec)
{
    register int i;

    if (!rk)
	return 0;

    rk->node_limit = rk->mem_limit * 1024 / sizeof(tree_node_t);
#ifdef POOL_ALLOC
    /* get a bunch of nodes for starters      */
    rk->free_nodes = 
	(tree_node_t *) (rk->alloc)(NODE_ALLOCATION * sizeof(tree_node_t));
    if (rk->free_nodes == 0) {
	fprintf (stderr, "rk: cannot malloc() even %ld bytes.\n",
		 NODE_ALLOCATION * sizeof(tree_node_t));
	return 0;
    }
#endif
    rk->next_free = 0;

    /* set up root and pointers into model      */
    rk->CBuf[0] = rk->Buf[0] = rk->root = create_node(rk);
    rk->root->up = rk->root->next = 0;
    rk->root->count = 1;
    for (i = 1; i <= rk->maxk; i++)
	rk->Buf[i] = 0;
    if (!zero_vec) 
	zero_vec = zero_freq_template;
    for (i = 0; i < MAX_SET; ++i)
        rk->zero_freq[i] = zero_vec[i];

    if (prime_vec) {
	const char* p;
	rk->psize = strlen(prime_vec);
        for (p = prime_vec; *p; ++p)
	    if (!move_up (rk, *p))
	        break;
        last_n_of_2 (rk->context, rk->maxk, "", prime_vec);
        last_n_of_2 (rk->old_context, rk->maxk, "", prime_vec);
    }

    return 1;
}

/* =======================================================================
 * Name - make_a_prediction
 *
 * Purpose -
 *
 * Arguments:
 *
 * Returns     function return --
 *
 *========================================================================
 */
int make_a_prediction (struct rk *rk, int num, char* partial_input)
{				/* sets s to the current prediction */
    context_str_t temp;
    char *a;
    int i;

    if (!rk || !partial_input)
	return 0;

    if (num < 0) num = 0;
    if (num > MAX_SET-1) num = MAX_SET-1;
    rk->pred_number = num;

    last_n_of_2 (temp, rk->maxk, rk->context, partial_input);

    if (strcmp (rk->old_context, temp) != 0)
	rk->pred_number = 0;
    strcpy (rk->old_context, temp);

    for (i = 1; i <= rk->maxk; i++)
	rk->CBuf[i] = 0;
    for (a = temp; *a; ++a)
	scan_up (rk, rk->CBuf, *a);
    find_first (rk);
    build_menu (rk);

    return 1;
}

/* =======================================================================
 * Name - move_up
 *
 * Purpose - Darkest of all JJD magic -- DJP
 *           Add character to uplist of given node, keeping uplist
 *           sorted in descending order.  Sweep through all nodes in list
 *           updating model.
 *
 * Arguments:
 *
 * Returns     function return --
 *
 *========================================================================
 */
static int move_up (struct rk *rk, int c)
{
    tree_node_t *xptr, *fxptr, *last_ptr, *last_fptr;
    enum {START, SCANNING, FOUND, END} state;
    tree_node_t *nptr;
    tree_node_t *move_up_result;
    int i;

    if (rk->memoryIsFull)
	return 0;
    rk->CharsInModel++; /* partially inserted chars will also be counted... */
    for (i = rk->maxk; i > 0; i --) {
	nptr = rk->Buf[i - 1];
	if (nptr == 0) {	/* no uplist */
	    move_up_result = 0;
	    goto END_MOVE_UP;
	}
	if (nptr->up == 0) {	/* new uplist */
	    state = START;
	    last_ptr = xptr = 0; /* pacify lint */
	} else {
	    fxptr = xptr = nptr->up;
	    last_fptr = last_ptr = 0;
	    state = SCANNING;
	    do {
		if (xptr == 0) /* end of uplist */
		    state = END;
		else {
		    if (xptr->count < fxptr->count) { /* remember insertion position */
			last_fptr = last_ptr;
			fxptr = xptr;
		    }
		    if (xptr->value == c) { /* found this character in uplist */
			state = FOUND;
			if (fxptr != xptr) {
			    if (last_fptr == 0) /* first entry in uplist */
				nptr->up = xptr;
			    else /* move entry in uplist */
				last_fptr->next = xptr;
			    last_ptr->next = xptr->next;
			    xptr->next = fxptr;
			}
		    } else {
			/* remember this link in chain and chain forward */
			last_ptr = xptr;
			xptr = xptr->next;
		    }
		}
	    } while (state == SCANNING);
	}
	switch (state) {
	case FOUND:
	    (xptr->count) ++;
	    if (xptr->count < rk->max_freq)
		break;
	    rk->FrequencySplits ++;
	    fxptr = nptr->up;
	    /*  last_fptr = fxptr; */
	    while (fxptr != 0) { /* JJD 9-86 */
#if 0
		/* Some forgetting of old nodes may take place here, but
		   not in this version.  It is quite complicated so for
		   simplicity we will forget it for now. */

		fxptr->count >>= 1; /* Forgets on halving 1 */
		/* DOESN'T FORGET A NODES SUBTREE NODES YET SO DON'T DO IT */
		/* MAY HAVE TO INCREASE max_freq TO REDUCE HALVING & OVERHEAD*/
		if (fxptr->count == 0) {
		    last_fptr->next = fxptr->next;
		    write (1, "1/2 free\n", 9);
		    FREE (fxptr);
		    fxptr = last_fptr->next;
		}
		else {
		    last_fptr = fxptr;
		    fxptr = fxptr->next;
		}
#else
		fxptr->count++;	/* Increment to at least 2 */
		fxptr->count >>= 1; /* Flatten count */
		fxptr = fxptr->next;
#endif
	    }
	    break;

	case START:
	case END:
	    xptr = create_node(rk);
	    if (xptr == 0)	/* memoryIsFull */
		return 0;
	    xptr->value = c;
	    xptr->count = 1;
	    xptr->up = xptr->next = 0;
	    if (state == START)
		nptr->up = xptr;
	    else
		last_ptr->next = xptr;
	    break;

	default:;
	}
	move_up_result = xptr;
    END_MOVE_UP:
	rk->Buf[i] = move_up_result;
    }
    return 1;
}

/* =======================================================================
 * Name - prime_from_file
 *
 * Purpose -
 *
 * Arguments:
 *
 * Returns     function return -- Keyboard status
 *
 *========================================================================
 */
int prime_from_file (struct rk *rk, char* filename)
{				/* JJD 9-86 */
    FILE *from;
    int c;

    if (!rk)
	return 0;

    if (rk->memoryIsFull) {
	return 0;
#if 0
	printf("But memory is full!  Can't expand predictive model.\n");
	fflush(stdout);
	goto ALL_DONE;
#endif
    }

    if ((from = fopen(filename, "r")) == 0) {
	fprintf (stderr, "rk: cannot open %s: %s\n", 
		 filename, strerror(errno));
	return 0;
    }
    do c = getc(from); while (c != EOF && move_up (rk, c));
    fclose(from);

    return 1;
}

/* =======================================================================
 * Name - scan_up
 *
 * Purpose - Auxiliary JJD magic -- DJP
 *           Find given character in uplist of given node  Sweep through
 *           all nodes in list updating model.
 *
 * Arguments:
 *
 * Returns     function return --
 *
 *========================================================================
 */
static void scan_up (struct rk *rk, context_vec_t theList, int c)
{
    tree_node_t * nptr;
    tree_node_t * xptr;
    int j;

    for (j = rk->maxk; j >= 1; j--) {
	nptr = theList[j - 1];
	if (nptr == 0) {
	    xptr = 0;
	    goto END_SCAN_UP;
	}
	xptr = nptr->up;
	for (;;) {
	    if (xptr == 0) {
		goto END_SCAN_UP;
	    }
	    if (xptr->value == c)
		goto END_SCAN_UP;
	    xptr = xptr->next;
	}
    END_SCAN_UP:;
	theList[j] = xptr;
    }
}

/* =======================================================================
 * Name - show_used_nodes
 *
 * Purpose -
 *
 * Arguments:
 *
 * Returns     function return -- Keyboard status
 *
 *========================================================================
 */
int show_used_nodes(struct rk *rk)
{
    if (!rk)
	return 0;

    fprintf(stderr,
	    "\r\n-o (order of model) %d, -f %d "
	    "(max freq, %ld actual splits)\r\n",
	    rk->maxk, rk->max_freq, rk->FrequencySplits);
    fprintf(stderr,
	    "%ld bytes each node, %ld nodes actually used, "
	    "heap chunks are %ld bytes\r\n",
	    sizeof(tree_node_t), rk->AllocationCount, 
	    NODE_ALLOCATION * sizeof(tree_node_t));
    fprintf(stderr,
	    "Nodes use over %ldK total bytes. Limit: %dK bytes.  "
	    "Already full: %s\r\n",
	    (rk->AllocationCount * sizeof(tree_node_t)) / 1024, rk->mem_limit,
	    rk->memoryIsFull ? "yes" : "no");
    fprintf(stderr,
	    "pred_number = %d, %ld characters in predictive model\r\n",
	    rk->pred_number, rk->CharsInModel);
    return 1;
}

/* =======================================================================
 * Name - update_the_model
 *
 * Purpose -
 *
 * Arguments:
 *
 * Returns     function return --
 *
 *========================================================================
 */
int update_the_model(struct rk *rk, char *s)
{				/* adds s into the model & updates log */
    context_str_t temp;

    if (!rk || !s)
	return 0;
    
    /* If predictive model is full, we're done. */
    if (rk->memoryIsFull)
	return 0;

    /* Build new context */
    strcpy (rk->context, last_n_of_2 (temp, rk->maxk, rk->context, s));

    /* Apply new context to model */
    while (*s != '\0' && move_up (rk, *s))
	s++;
    return 1;
}


/* last_n_of_2 -- given strings a and b return the last n chars of ab */
static char* last_n_of_2 (char* out, int n, const char* a, const char* b)
{
    int alen, blen;
    
    if (!out)
	return 0;
    
    /* assert (n < sizeof out); */

    if (!a || !b) {
	out[0] = '\0';
    } else {
	blen = strlen (b);
	if (n <= blen) {
	    strcpy (out, b + blen - n);
	} else {
	    alen = strlen (a);	
	    n -= blen;
	    if (n <= alen)
		strcpy (out, a + alen - n);
	    else
		strcpy (out, a);
	    strcat (out, b);
	}
    }
    return out;
}
